/* Quick test of VLSI PID's interrupt handling system
 * Sets up an interrupt handler that waits for the 'panic' button
 * on the PID to be pressed and then prints a message.
 */

#include <ARM/arm.h>
#include <ARM/vy86pid.h>
#include <ARM/trap.h>
#include <intr.h>
#include <thread.h>
#include <root.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>

IntrHandler 	e;

#pragma no_check_stack	/* MUST be used for C based interrupt handlers */

bool
PanicHandler(
	     void *	pSemaphore,
	     int	vec )
{
  /* Reset interrupt */
  hw_INTC->IRQRST = hw_intc_resetirq_panic;
  
#if 0
  /* Calling HardenedSignal directory goes via the trap interface, which is slower */
  HardenedSignal( (Semaphore *)pSemaphore );
#else
  PseudoTrap( (word)pSemaphore, 0, 0, TRAP_HardenedSignal);
#endif
  
  return TRUE;
}

#pragma check_stack

static void terminate(int sig)
{
  printf("\nPidpanic terminated\n");

  RemIntrHandler(&e);

  /* Terminate the program */
  exit(0);

}

int
main( void )
{
  Semaphore	psem;
  RootStruct *	rs;

  /* Set up ctrl-c handling */
  struct sigaction New;
  New.sa_handler = terminate;
  New.sa_mask    = 0;
  New.sa_flags   = SA_SETSIG | SA_ASYNC;
  sigaction( SIGINT, &New, NULL );

  rs = GetRoot();

  e.Pri = StandardPri;
#ifdef __VY86PID
  e.Id = INTR_IRQ_PANIC;
#endif

  e.Code = PanicHandler;
  e.Data = &psem;

  SetIntrHandler( &e );
  
  /* Not required as PANIC button interrupt is always enabled. */
  /*	hw_INTC->IRQM = rs->IRQM_softcopy |= hw_intc_enableirq_panic;*/
  
  InitSemaphore( &psem, 0 );

#ifdef __VY86PID
  {
    int	iOldMode = ARM_GetMode();

    /* Enter service mode to get mask value */
    ARM_SwitchMode( SVCmodeG );

    printf("pid mask %lx, status %x\n", rs->IRQM_softcopy , hw_INTC->IRQS);

    /* Then return to original mode */
    ARM_SwitchMode( iOldMode );
  }
#endif

  while (TRUE)
    {
      printf("Waiting for Panic to be pressed\n");

      Wait(&psem);

      printf("Panic WAS pressed.\n");
    }
}
