/*> undef.c <*/
/*---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/signal.h> /* we need the Helios extensions */

/*---------------------------------------------------------------------------*/
/* The "_word()" pseudo-function provides a means to insert 32bit
 * values directly into the code stream. Unfortunately without any
 * register binding checks. However, along with the assumption that
 * the C compiler will generate APCS conforming functions we can
 * provide a simple example without resorting to a seperate piece of
 * assembler source.
 */
unsigned int gen_undefined_instruction(int a1)
{
 _word(0xE6000010) ; /* our new instruction */
 return(a1) ;
}

/*---------------------------------------------------------------------------*/

void normal_handler(int sig)
{
 if (sig != SIGSEGV)
  {
   printf("normal_handler: unrecognised signal\n") ;
   exit(1) ;
  }

 printf("normal_handler: continuing the thread\n") ;
 return ;
}

/*---------------------------------------------------------------------------*/

/* The following includes are required when dealing with lo-level exceptions */
#include <helios.h>
#include <cpustate.h>

/* NOTE: This handler accepts a new (sigstate *) parameter */
void undef_handler(int sig,sigstate *state)
{
 printf("undef_handler entered: sig %d, sigstate = 0x%08X\n",sig,(int)state) ;

 /* The "state" is NULL for signals caused by "raise". It is only
  * non-null for asynchronous low-level system exceptions.
  */
 if ((sig == SIGILL) && (state != NULL))
  {
   SaveState *sstate = (SaveState *)(state->threadstate) ;
   printf("state->subcode = %d\n",state->subcode) ;
   printf("state->data = 0x%08X\n",(int)(state->data)) ;
   printf("state->threadstate = 0x%08X\n",(int)sstate) ;
   printf("threadstate:\n") ;
   printf("  PC = 0x%08X\n",(int)(sstate->CPUcontext.R_PC)) ;
   printf("  instruction at PC = 0x%08X\n",*((unsigned int *)(sstate->CPUcontext.R_PC))) ;

   if (*((unsigned int *)(sstate->CPUcontext.R_PC)) != 0xE6000010)
    {
     printf("undef_handler: unrecognised instruction\n") ;
     exit(1) ;
    }

   printf("  a1 = 0x%08X (%d)\n",(int)(sstate->CPUcontext.R_A1),(int)(sstate->CPUcontext.R_A1)) ;

   sstate->CPUcontext.R_A1 += 1000 ; /* perform our operation */
   sstate->CPUcontext.R_PC += 4 ; /* number of bytes in the instruction */
   /* We have now "directly" updated the state which the thread will
    * be re-started with.
    */
  }
 else
  {
   printf("terminating task\n") ;
   exit(1) ;
  }

 printf("undef_handler: continuing the thread\n") ;
 return ;
}

/*---------------------------------------------------------------------------*/

int main(int argc,char **argv)
{
 unsigned int result ;

 signal(SIGILL, (void(*)(int)) undef_handler) ;
 signal(SIGSEGV,normal_handler) ;

 printf("About to raise SIGSEGV\n") ;
 raise(SIGSEGV) ;
 printf("raise(SIGSEGV) returned\n\n") ;

 printf("About to generate undefined instruction (a1 = 1969)\n") ;
 result = gen_undefined_instruction(1969) ;
 printf("result = %d\n",result) ;

 return(0) ;
}

/*---------------------------------------------------------------------------*/
/*> EOF undef.c <*/
