/*------------------------------------------------------------------------
--                                                                      --
--			H E L I O S   S E R V E R S			--
--			---------------------------			--
--                                                                      --
--             Copyright (C) 1991, Perihelion Software Ltd.             --
--                        All Rights Reserved.                          --
--                                                                      --
-- romserver.c								--
--                                                                      --
--	This is a ROM file server derived from the Helios include       --
--      server example documented in the Helios Parallel Operating      --
--      system book. Reference to the example is recomended before      --
--      modification is made to this example.                           --
--                                                                      --
--	Author:  BLV 21.3.91						--
--      Author:  ACC 16.5.91                                            --
--	Author:  TC  10.12.93						--
--									--
--  AHC says ...							--
--	Rom server modified to pass the loader the address of any 	--
-- programs in the rom disk directly when they are in rom or ram (ie	--
-- when the rom disk file is either read into memory or already in	--
-- memory).								--
--									--
------------------------------------------------------------------------*/

/* $Id: romserver.c,v 1.10 1995/03/27 10:27:08 tony Exp $ */

#include <helios.h>
#include <syslib.h>
#include <servlib.h>
#include <task.h>
#include <sys/types.h>
#include <fcntl.h>
#include <posix.h>
#include <string.h>

#include <nonansi.h>	/* for Fork () */

extern word Timer (void);

#include <root.h>

/* How to recognize a ROM disk - defined in module.h in releases later than 1.3.1 */
#ifndef T_RomDisk
# define	T_RomDisk	0x60fb60fbl
#endif

#if 0
#define ROMSERVER_DEBUG
#endif

#ifdef __C40
/*
 * Define this if you don't want the loader to use eXecute In Place for
 * programs in the ROMdisk.  Mainly for C40's, where the executables in
 * the ROMdisk are too far away from the C Base.
 */
#ifndef NO_XIP
#define NO_XIP
#endif

#endif

#pragma -s1

/*
 * Define FILEBASED means that the filing system is being provided by a file
 * on another server. For example a host development computer which holds an
 * image of the filing system on disk. 
 * In this case the definition FILEBASEDFILE is required.
 * All reads on this filing system will be modified and passed onto that server
 * so that that reply goes straight back to the client instead of through
 * here.
 * 
 * Define ROMBASED means that the filing system is in local memory. This can
 * be done in either of two ways :-
 * 1) By copying a remote file into memory - In which case a definition of 
 *     ROMBASEDFILE is required.
 * 2) By finding the nucleus vector table and checking all the entries
 *     until T_RomDisk is found - In which case ROMBASEDNUCLEUS is required.
 */

/*
 * We shall access the ROM using the following generic functions 
 * 
 * ROMINFO * OpenRom();
 *
 * int ReadRom (ROMINFO *	stream,
 *		BYTE *		buffer,
 *		WORD		size,
 *		WORD		timeout);
 *
 * void CloseRom (ROMINFO *	stream);
 */

/* Set default to ROMBASED */
#ifndef FILEBASED
#  ifndef ROMBASED
#      define ROMBASED
#  endif
#endif


/*
 *		SPECIFIC STRUCTURE / DEFINITIONS (dependent on type of romdisk)
 */

#ifdef FILEBASED

/* default file holding romdisk - romdir.fil in current directory */
#ifndef FILEBASEDFILE
# define FILEBASEDFILE "romdir.fil"
#endif

typedef Stream *	ROMHANDLE;

static struct ROMINFO *	OpenRom (void);

#define ReadRom(rominfo,buf,size,timeout) \
		Read (rominfo->instream, buf, size, timeout)

#define CloseRom(rominfo) \
		Close (rominfo->instream); Free (rominfo)

#else	/* ROMBASED */

/* set default to ROMBASEDFILE - default file is romdir.fil in current directory */
# ifndef ROMBASEDFILE
#  ifndef ROMBASEDNUCLEUS
#    define ROMBASEDFILE "romdir.fil"
#  endif
#endif

typedef struct RomFile
{
    MPtr	rombuffer;
    WORD 	pos;
} RomFile;

#define ROMHANDLE	RomFile

static struct ROMINFO *	OpenRom(void);

static int	ReadRom (struct ROMINFO *	stream,
			 BYTE * 		buffer,
			 WORD 			size,
			 WORD 			timeout);

static int	CloseRom (struct ROMINFO * stream);

#endif /* FILEBASED */


/*
 *		FUNCTION DELCARATIONS
 */

static void	do_open (ServInfo *);
static void	do_delete(ServInfo *);
static void	do_serverinfo(ServInfo *);


/*
 *		GENERIC STRUCTURES / DEFINITIONS
 */

typedef struct ROMINFO
{
#ifndef ROMBASEDNUCLEUS
    struct ImageHdr	ihdr;		/* image header is only in files */
#endif
    struct Module	modhdr;
    long		romsize;
    ROMHANDLE	instream;
} ROMINFO;

ROMINFO *	RomInfo;

static DirNode	Root;

static char	ProcessorName[IOCDataMax];

static DispatchInfo RomDispatchInfo =
{
    &Root,
    NullPort,
    SS_RamDisk,
    NULL,
    {NULL, 0},
    {
	{do_open, 2000},	/* FG_Open		*/
	{InvalidFn, 2000},	/* FG_Create		*/
	{DoLocate, 2000},	/* FG_Locate		*/
	{DoObjInfo, 2000},	/* FG_ObjectInfo	*/
	{do_serverinfo, 2000},	/* FG_ServerInfo	*/
	{do_delete, 2000},	/* FG_Delete		*/
	{DoRename, 2000},	/* FG_Rename		*/
	{DoLink, 2000},		/* FG_Link		*/
	{DoProtect, 2000},	/* FG_Protect		*/
	{DoSetDate, 2000},	/* FG_SetDate		*/
	{DoRefine, 2000},	/* FG_Refine		*/
	{NullFn, 2000},		/* FG_CloseObj		*/
	{DoRevoke, 2000},	/* FG_Revoke		*/
	{InvalidFn, 2000},	/* Reserved 		*/
	{InvalidFn, 2000}	/* Reserved		*/
    }
};

/*
 *		CODE / FUNCTIONS
 */

static WORD	TellRom ()
{
    return (RomInfo->instream.pos);
}

static void	SeekRom (WORD	p)
{
    RomInfo->instream.pos = p;
}

#ifdef FILEBASED

static ROMINFO * OpenRom (void)
{
    ROMINFO *	r;
    Object *	rom_disk;
    
    r = New (ROMINFO);
    if (r == NULL)
    {
	IOdebug ("rom: failed to allocate ROMINFO structure");
	
	return NULL;
    }

    rom_disk = Locate (Null(Object), FILEBASEDFILE);
    if (rom_disk == Null(Object))
    {
	IOdebug ("rom : failed to locate %s", FILEBASEDFILE);

	return NULL;
    }

    r->instream = Open (rom_disk, Null(char), O_ReadOnly);
    if (r->instream == Null(Stream))
    {
	IOdebug ("rom : failed to open %s, fault %x", rom_disk->Name,
		 		Result2 (rom_disk));

	return NULL;
    }

    Close (rom_disk);

    return r;
}

#else	/* ROMBASED */

static ROMINFO * OpenRom (void)
{
    ROMINFO *	r = New (ROMINFO);

#ifdef ROMBASEDFILE
    Object *	rom_disk;
    Stream *	f;
    WORD 	size;
    WORD	n;

    /* attempt to find rom disk globally */
    rom_disk = Locate (Null(Object), ROMBASEDFILE);
    if (rom_disk == Null(Object))
    {
	/* attempt to find it locally */
	rom_disk = Locate (CurrentDir, ROMBASEDFILE);

	if (rom_disk == Null (Object))
	{
	    IOdebug ("rom : failed to locate %s", ROMBASEDFILE);

	    return NULL;
	}
    }

    f = Open (rom_disk, Null(char), O_ReadOnly);
    if (f == Null(Stream))
    {
	IOdebug ("rom : failed to open %s, fault %x", rom_disk->Name,
		 		Result2 (rom_disk));

	return NULL;
    }

    Close (rom_disk);

    if ((size = GetFileSize (f)) < Err_Null)
    {
	IOdebug ("rom : failed to get file size of %s, fault %x",
		 		f->Name, Result2 (f));

	return NULL;
    }

    r->instream.rombuffer = (MPtr)(Malloc (size));
    if (r->instream.rombuffer == (MPtr)NULL)
    {
	IOdebug ("rom: failed to allocate memory for ROM buffer");

	Close (f);

	return NULL;
    }

#ifdef ROMSERVER_DEBUG
    IOdebug ("rombuffer starts at 0x%x (size: %d bytes)", 
	     		(long)(r->instream.rombuffer), size);
#endif

    n = Read (f, (BYTE *) r->instream.rombuffer, size, -1);
    if (n != size)
    {
	IOdebug("rom : failed to get %d bytes from %s, fault %x",
			size, f->Name, Result2 (f));

	return NULL;
    }

    Close (f);

#else	/* ROMBASEDNUCLEUS */

    MPtr 	vector_ptr = MInc_(GetSysBase (), sizeof (MPtr));
    MPtr	p = 0;

#ifdef ROMSERVER_DEBUG
    IOdebug ("Nucleus Base  = %x", vector_ptr);
    IOdebug ("*Nucleus Base = %x", MWord_(vector_ptr, 0));
#endif

    /*
     * Find the romdisk in the nucleus by looking at all entries
     * in the vector table and examining their type for T_RomDisk
     */

    while (MWord_(vector_ptr, 0))
    {
	p = MRTOA_(vector_ptr);

#ifdef ROMSERVER_DEBUG
	IOdebug ("vector_ptr =  %x", vector_ptr);
	IOdebug ("*vector_ptr = %x", MWord_(vector_ptr, 0));
	
	IOdebug ("p  = %x", (long)p);
	IOdebug ("*p = %x", MWord_(p, 0));
	IOdebug ("ModuleWord_(p, Type) = %x", ModuleWord_(p, Type));
#endif
	if (ModuleWord_(p, Type) == T_RomDisk)
	{
	    break;
	}
	vector_ptr = MInc_(vector_ptr, sizeof (MPtr));
    }

    if (MWord_(vector_ptr, 0) == NULL)
    {
	IOdebug ("Failed to find rom disk in nucleus");
	
	return NULL;
    }

    r->instream.rombuffer = p;

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: OpenRom () - rombuffer starts at %x",
	     		(long)(r->instream.rombuffer));
#endif

#endif	/* ROMBASEDFILE */

    r->instream.pos = 0;

    return r;
}

static int ReadRom (ROMINFO * 	stream,
		    BYTE * 	buffer,
		    WORD 	size,
		    WORD 	timeout)
{
    MPtr	mp = stream->instream.rombuffer;
    word	offset = stream->instream.pos;

    timeout = timeout;	/* keep the compiler happy */

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: ReadRom () - reading %d bytes from offset %d (0x%x)", size, offset, offset);
    IOdebug ("rom: ReadRom () - mp     = %x", (long)(mp));
    IOdebug ("rom: ReadRom () - offset = %d", offset);

    IOdebug ("rom: ReadRom () - current word is %x", MWord_(mp, offset));
#endif

    MData_(buffer, mp, offset, size);

    stream->instream.pos += size;

    return (int)size;
}

static int CloseRom (ROMINFO *	stream)
{
#ifdef ROMBASEDFILE
    Free (stream->instream.rombuffer);
#endif
    Free(stream);

    return NULL;
}

#endif	/* ROMFILEDBASED */

static ObjNode * getobjnode()
{
    ObjNode *	objnode = New (ObjNode);

    if (ReadRom (RomInfo, (BYTE *)objnode, (WORD)(sizeof (ObjNode)), -1) != sizeof (ObjNode))
    {
    	IOdebug ("rom: error when reading objnode");
    	Exit (256);
    }

    RomInfo->romsize += sizeof (ObjNode);

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: getobjnode () - name: %s, type: 0x%x", objnode->Name, objnode->Type);
#endif

    return objnode;
}

#ifdef NO_XIP
#define LoadRomImage(n,a)
#else
static void LoadRomImage (char *	name,
			  MPtr		addr)
{
    static Object *	loader = Null (Object);
    MCB		mcb;

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: LoadRomImage (%s, %x)", name, (int)addr);
#endif

    if (loader == Null (Object))
    {
	while ((loader = Locate (NULL, "/loader")) == NULL)
	{   /* find the first time 'round */
	    ;
	}
    }

    mcb.Data    = (byte *)(Malloc (IOCDataMax));
    mcb.Control = (word *)(Malloc (sizeof (IOCCommon) + (2 * sizeof (word))));

    InitMCB (&mcb, 0, MyTask->IOCPort, NullPort, FC_GSP | FG_RomCreate);

    MarshalCommon (&mcb, loader, NULL);
    MarshalString (&mcb, name);
    MarshalWord (&mcb, (word)addr);

#ifdef OLD_VERSION
    PutMsg (&mcb);
#else
    if (PutMsg (&mcb) != Err_Null)
    {
	/* Try 5 more times, then give up */
	int	retries = 5;

	while (retries-- != 0)
	{
	    if (PutMsg (&mcb) == Err_Null)
	    {
		break;
	    }

	    if (retries == 0)
	    {
		IOdebug ("Failed to send message to /loader");

		return;
	    }
	}
    }
#endif /* OLD_VERSION */

    Free (mcb.Control);
    Free (mcb.Data);

    Close (loader);

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: LoadRomImage (%s, %x) finished", name, (int)addr);
#endif
}
#endif /* NO_XIP */

/*
 * extract_files () -
 *	Initialises the directory structure
 */
static void extract_files (DirNode * 	parent,
			   WORD		number)
{
    ObjNode *	objnode;

#ifdef ROMSERVER_DEBUG
    IOdebug ("extract_files (%s, %d)", parent->Name, number);
#endif

    while (number-- > 0)
    {
    	WORD	number_entries;

    	word	objsize;

    	objnode = getobjnode ();

    	if (objnode->Size == 0)
    	{
    	    IOdebug ("extract_files () - found a zero sized object");
    	}

    	if (   (objnode->Type != Type_File)
    	    && (objnode->Type != Type_Directory))
    	{
    		IOdebug ("rom : ROM image appears to be corrupt, found %x for objnode->Type", objnode->Type);

    		IOdebug ("rom : ROM image appears to be corrupt");

    		Exit (256);
    	}

    	number_entries = objnode->Size;

    	/*
    	 * The following call to InitNode () sets the Size
    	 * of the objnode to 0, thus seriously screwing up the
    	 * world later on.
    	 *
    	 * The following is a fix to get around this problem, but
    	 * there are two questions -
    	 * 	1. In a ROM world, how is it able to modify the value
    	 *	   at all?
    	 *	2. Why doesn't this screw up on any other processor?
	 *
	 * Stranger and stranger!  The setting of the size to 0 only
	 * effects files, not directories.
    	 */
    	objsize = objnode->Size;

    	InitNode (objnode,
    		  objnode->Name,
    		  (int)(objnode->Type),
    		  (int)(objnode->Flags),
    		  objnode->Matrix);

	if (objnode->Type != Type_Directory)
	{
	    objnode->Size = objsize;
	}

    	Insert(parent, objnode, FALSE);

#ifdef ROMSERVER_DEBUG
	IOdebug ("extract_files (): found objnode ...");
	IOdebug ("                      Name: %s", objnode->Name);
	IOdebug ("                      Type: %x", objnode->Type);
	IOdebug ("                      Size: %x", objnode->Size);
#endif

    	if (objnode->Type == Type_Directory)
    	{
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: extract_files () - found DirNode %s",
    			    	objnode->Name);
#endif

    		/* If a directory, recurse down to all its entries */
    		InitList (&((DirNode *)objnode)->Entries);

    		extract_files ((DirNode *)objnode, number_entries);
    	}
    }
}

static void extract_executables (WORD		number)
{
    ObjNode *	objnode;

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: extract_executables (%d)", number);
#endif

    while (number-- > 0)
    {
    	WORD	number_entries;

    	objnode = getobjnode ();

    	if (   (objnode->Type != Type_File)
    	    && (objnode->Type != Type_Directory))
    	{
    		IOdebug ("rom : ROM image appears to be corrupt, found %x for objnode->Type", objnode->Type);

    		IOdebug ("rom : ROM image appears to be corrupt");

    		Exit (256);
    	}

    	number_entries = objnode->Size;

    	if (objnode->Type == Type_Directory)
    	{
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: extract_executables () - found DirNode %s",
    			    	objnode->Name);
#endif
    		extract_executables (number_entries);
    	}
    	else
    	{
    		MPtr	addr;
    		word	offset;

#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: extract_executables () - found ObjNode %s",
    				objnode->Name);
#endif

    		/*
    		 * Address of the executable (if such it is) is
    		 *	address of buffer + offset
    		 * where offset can be found in the first word of the
    		 * Contents field in objnode.
    		 */

    		offset = (word)((&objnode->Size)[1]);
#ifndef ROMBASEDNUCLEUS
    		/*
    		 * Real rom disk files have an Image Header placed
    		 * on the front which the offset does not take into
    		 * account.
    		 */
    		offset += sizeof (ImageHdr);
#endif
    		addr = MInc_(RomInfo->instream.rombuffer, offset);

#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: %s appears at %x in romdisk",
    				objnode->Name, (long)addr);
    		IOdebug ("rom: MWord_(addr, 0) = %x", MWord_(addr, 0));
#endif

    		/* If the file is an executable, enter it into the loader */
    		if (MWord_(addr, 0) == Image_Magic)
    		{
#ifdef ROMSERVER_DEBUG
    			IOdebug ("rom: %s is an executable",
    					objnode->Name);

#define MyMWord_(a,n)	MWord_(a, n*4)

    			IOdebug ("rom: 1st 16 words are - ");
    			IOdebug ("	%x %x %x %x",
    				MyMWord_(addr, 0), MyMWord_(addr, 1),
    				MyMWord_(addr, 2), MyMWord_(addr, 3));
    			IOdebug ("	%x %x %x %x",
    				MyMWord_(addr, 4), MyMWord_(addr, 5),
    				MyMWord_(addr, 6), MyMWord_(addr, 7));
    			IOdebug ("	%x %x %x %x",
    				MyMWord_(addr, 8), MyMWord_(addr, 9),
    				MyMWord_(addr, 10), MyMWord_(addr, 11));
    			IOdebug ("	%x %x %x %x",
    				MyMWord_(addr, 12), MyMWord_(addr, 13),
    				MyMWord_(addr, 14), MyMWord_(addr, 15));
#endif

    			LoadRomImage (objnode->Name, addr);
    		}
    		objnode->Size = number_entries;
    	}
    }
}

#ifdef ROMBASED

/*
 * Fork'ed function - waits for /helios/lib to appear, and then calls
 * extract_executables ().
 */
/*
static void waitfor_helios (DirNode *	parent,
			    WORD	number)
*/
static void waitfor_helios (WORD	number)
{
    int	retries = 0;

    char 	rootname[IOCDataMax];

    Object *	heliosdir;

#if 0
    Object *	libdir;		/* Debug /helios ROMdisk only */
#endif

    if (Root.Name[0] != '/')
    {
	rootname[0] = '/';
	rootname[1] = '\0';
    }
    else
    {
	rootname[0] = '\0';
    }
    strcat (rootname, Root.Name);

    /*
     * Wait for Root.Name to appear (means that the ROMdisk server is
     * running).
     */

#ifdef ROMSERVER_DEBUG
    IOdebug ("waitfor_helios (): waiting for %s", rootname);
#endif

/*    while ((heliosdir = Locate (Null(Object), "/helios")) == Null(Object)) */
    while ((heliosdir = Locate (Null (Object), rootname)) == Null (Object))
    {

#if (defined __TRAN || defined __C40)
	if (retries > 10000)
	{
	    IOdebug ("Failed to find /helios after 10000 attempts");

	    return;
	}
#else
	if (retries > 15)
	{
	    IOdebug ("Failed to find /helios after 15 attempts");

	    return;
	}
	
	Delay (OneSec / 1000L);	/* wait for 1 milli second */
#endif
	retries++;
    }
				
#if 0
    while ((libdir = Locate (heliosdir, "lib")) == NULL)
    {
#ifdef __TRAN
	if (retries > 10000)
	{
	    IOdebug ("Failed to find /helios/lib after 10000 attempts");

	    return;
	}
#else
	if (retries > 15)
	{
	    IOdebug ("Failed to find /helios/lib after 15 attempts");

	    return;
	}

	Delay (OneSec / 1000L);
#endif
	retries++;
    }
#endif

#ifdef ROMSERVER_DEBUG
    IOdebug ("waitfor_helios (): located %s", rootname);
#endif

    extract_executables (number);

    {
    	MCB m;

#ifdef ROMSERVER_DEBUG
	IOdebug ("rom: sending 456 message to parent ... ");
#endif
    	InitMCB (&m, 0, MyTask->Parent, NullPort, 0x456);

    	if (PutMsg (&m) != Err_Null)
	{
	    IOdebug ("rom: failed to send 456 message");
	}
	
#ifdef ROMSERVER_DEBUG
	IOdebug ("rom: successfully sent 456 message");
#endif 
    }
}

#define STACKSIZE	0x8000

int call_extract_executables (WORD	number)
{
    int		r;

    r = Fork (STACKSIZE,
          waitfor_helios,
          sizeof (WORD),
          number);

    return r;
}
#endif

#ifdef USE_getrombuffer

static void getrombuffer (ROMINFO *	f)
{
    WORD size;

    if (ReadRom (f, (BYTE *)(&size), 4L, -1) != 4)
    {
    	IOdebug ("failed to read rom buffer size");
    	Exit (256);
    }

    f->romsize = size + 4;
}
#endif

/* 
 *		SERVER UTILITY FUNCTIONS
 */

static void handle_read (MCB *		m,
			 ObjNode * 	f)
{
    ReadWrite	*rw	= (ReadWrite *)(m->Control);
    WORD		 pos	= rw->Pos;
    WORD		 size	= rw->Size;
    WORD		 offset	= (&f->Size)[1];
    WORD		 i;
    Port		 reply;
    WORD		 sequence;
    word		 objsize = f->Size;

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom:    handle_read () - read %d bytes from position %d in file %s", size, pos, f->Name);
#endif

    if (pos < 0)
    {
    	ErrorMsg (m, EC_Error + EG_Parameter + 1);

#ifdef ROMSERVER_DEBUG
    	IOdebug ("rom:    handle_read () - pos (%d) < 0", pos);
#endif

    	return;
    }

    if (size < 0)
    {
    	ErrorMsg (m, EC_Error + EG_Parameter + 2);
#ifdef ROMSERVER_DEBUG
    	IOdebug ("rom:    handle_read () - size (%d) < 0", size);
#endif

    	return;
    }

    if (pos >= objsize)
    {
    	m->MsgHdr.FnRc = ReadRc_EOF;

#ifdef ROMSERVER_DEBUG
    	IOdebug ("rom:    handle_read () - pos (%d) >= objsize (%d)", pos, objsize);
#endif
    	ErrorMsg (m, 0);

    	return;
    }

    if ((pos + size) > objsize)
    {
    	size = objsize - pos;
    	rw->Size = size;
    }

    if (size == 0)
    {
    	m->MsgHdr.FnRc = ReadRc_EOD;

#ifdef ROMSERVER_DEBUG
    	IOdebug ("rom:    handle_read () - size == 0");
#endif
    	ErrorMsg (m, 0);
    	return;
    }

    /* For file-based pseudo-romdisks, the read request should	*/
    /* be forwarded directly to the file server unmodified.		*/
#ifdef FILEBASED
    m->MsgHdr.Dest	= RomInfo->instream->Server;

    rw->Pos += offset;

    PutMsg(m);
#else
    /* For real ROM systems the data can be sent directly. It may	*/
    /* have to be split over more than one message.			*/
    reply	  = m->MsgHdr.Reply;
    sequence = 0;

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom:    handle_read () - size = %d", size);
#endif

#define BUF_LEN		65536	/* Size of buffer				  */
#define BUF_LEN_FILL	65532	/* How many bytes to put in buffer, word aligned! */

    for (i = 0; i < size; i += (BUF_LEN_FILL))
    {
    	m->MsgHdr.Dest	= reply;
    	m->MsgHdr.Reply	= NullPort;
    	m->MsgHdr.ContSize	= 0;

    	/* Is this the last message ? */
    	if ((i + (BUF_LEN_FILL)) < size)
    	{	/* Not yet */
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom:    handle_read () - not last message");
#endif
    		m->MsgHdr.DataSize	= (BUF_LEN_FILL);
    		m->MsgHdr.Flags	= MsgHdr_Flags_preserve;
    		m->MsgHdr.FnRc	= sequence + ReadRc_More;
    		sequence += ReadRc_SeqInc;
    	}
    	else
    	{
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom:    handle_read () - last message");
#endif

    		m->MsgHdr.DataSize = (int)(size - i);
    		m->MsgHdr.Flags    = 0;
    		m->MsgHdr.FnRc     = sequence + ReadRc_EOD;
    	}

    	{
    		WORD   	j;
    		MPtr	buf = MInc_(RomInfo->instream.rombuffer, pos + offset + i);
    		static char	buffer[BUF_LEN];
#ifdef ROMSERVER_DEBUG
		IOdebug ("rom:    handle_read () - buf is 0x%x", (long)(buf));
		IOdebug ("rom:    handle_read () - pos is %d 0x%x", (long)(pos), (long)(pos));
		IOdebug ("rom:    handle_read () - offset is (now) %d 0x%x", (long)(offset), (long)(offset));
#ifdef __C40
		IOdebug ("rom:    handle_read () - 1st word is 0x%x", MWord_(buf, 0));
#else
    		IOdebug ("rom:    handle_read () - 1st 4 bytes are %x, %x, %x, %x", 
			 buf[0], buf[1], buf[2], buf[3]);
#endif /* __C40 */
#endif

    		/* Alenia bug fix. */
    		for(j = 0; j < m->MsgHdr.DataSize ; j++)
    		{
    			MData_(buffer + j, buf, j, 1);
    		}
		
    		m->Data = buffer;
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom:    handle_read () - 1st 4 bytes are %x, %x, %x, %x",
    					(m->Data)[0],
    					(m->Data)[1],
    					(m->Data)[2],
    					(m->Data)[3]);
#endif
    	}
    	if (PutMsg (m) < Err_Null)
    	{
    		IOdebug ("rom:    handle_read () - PutMsg () error");
    	}
    }
#endif	/* FILEBASED */
}

static void handle_seek (MCB *		m,
			 ObjNode * 	f)
{
    SeekRequest *	req = (SeekRequest *)(m->Control);

    WORD 	curpos = req->CurPos;
    WORD 	mode   = req->Mode;
    WORD 	newpos = req->NewPos;

    switch (mode)
    {
    case S_Beginning:
    	break;
    case S_Relative:
    	newpos += curpos;
    	break;
    case S_End:
    	newpos += f->Size;
    	break;
    }

    if (newpos > f->Size)
    	newpos = f->Size;

    if (newpos < 0)
    	newpos = 0;

    InitMCB (m, 0, m->MsgHdr.Reply, NullPort, Err_Null);
    MarshalWord (m, newpos);
    PutMsg (m);
}


/*
 *		SERVER FUNCTIONS
 */

static void do_open (ServInfo *	servinfo)
{
    MCB *		m = servinfo->m;
    MsgBuf *	r;
    DirNode *	d;
    ObjNode *	f;
    IOCMsg2 *	req = (IOCMsg2 *)(m->Control);
    Port 		reqport;
    BYTE *		data = m->Data;
    char *		pathname = servinfo->Pathname;

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: do_open () - pathname = %s", pathname);
#endif
    d = GetTargetDir (servinfo);
    if (d == Null(DirNode))
    {
    	IOdebug ("rom: do_open () - failed to find target dir %s", pathname);

    	ErrorMsg (m, Err_Null);

    	return;
    }

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: do_open () - found dir %s", d->Name);
#endif

    f = GetTargetObj (servinfo);
    if (f == Null(ObjNode))
    {
#ifdef ROMSERVER_DEBUG
    	IOdebug ("rom: do_open () - failed to find target obj %s", 
		 servinfo->Pathname);
#endif

    	ErrorMsg (m, Err_Null);
    	return;
    }

    if (f->Size == 0)
    {
        IOdebug ("handle_read () - Found object in ROMserver with zero size");
    }

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: do_open () - found file %s, size %d", f->Name, f->Size);
#endif

    if (!CheckMask (req->Common.Access.Access,
        		(AccMask)((req->Arg.Mode) & Flags_Mode)))
    {
#ifdef ROMSERVER_DEBUG
    	IOdebug ("rom: do_open () - CheckMask () failed");
#endif

    	ErrorMsg (m, EC_Error + EG_Protected + EO_File);

    	return;
    }

    r = New (MsgBuf);
    if (r == Null (MsgBuf))
    {
#ifdef ROMSERVER_DEBUG
    	IOdebug ("rom: do_open () - failed to get memory for return message");
#endif
    	ErrorMsg (m, EC_Error + EG_NoMemory + EO_Message);

    	return;
    }

    FormOpenReply (r,
    	       m,
    	       f,
     	       Flags_Server | Flags_Closeable | Flags_MSdos,
    	       pathname);

    reqport = NewPort ();
    r->mcb.MsgHdr.Reply = reqport;

    PutMsg (&r->mcb);

    Free (r);

    if (f->Type == Type_Directory)
    {
    	DirServer (servinfo, m, reqport);
    	FreePort (reqport);

#ifdef ROMSERVER_DEBUG
    	IOdebug ("rom: do_open () - %s is a directory with %d entries", f->Name, f->Size);
	IOdebug ("rom: do_open () - finished");
#endif
    	return;
    }

    f->Account++;

    UnLockTarget (servinfo);

    forever
    {
    	WORD e;

    	m->MsgHdr.Dest = reqport;
    	m->Timeout	 = StreamTimeout;
    	m->Data	 = data;

    	e = GetMsg (m);
    	if (e == EK_Timeout)
    	{
#ifdef ROMSERVER_DEBUG
	    IOdebug ("rom: do_open () - loop timeout, breaking");
#endif
	    break;
    	}
    	if (e < Err_Null)
	{
#ifdef ROMSERVER_DEBUG
	    IOdebug ("rom: do_open () - get message failed, continuing");
#endif
	    continue;
	}

    	Wait (&(f->Lock));

    	switch (m->MsgHdr.FnRc & FG_Mask)
    	{
    	case FG_Read:
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: do_open () - FG_Read");
#endif
    		handle_read (m, f);
    		break;

    	case FG_Close:
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: do_open () - FG_Close");
#endif
    		if (m->MsgHdr.Reply != NullPort)
    		{
    			ErrorMsg (m, Err_Null);
    		}
    		FreePort (reqport);
    		f->Account--;
    		Signal (&(f->Lock));

#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: do_open () - finished");
#endif

    		return;

    	case FG_GetSize:
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: do_open () - FG_GetSize");
#endif
    		InitMCB (m, 0, m->MsgHdr.Reply, NullPort, Err_Null);
    		MarshalWord (m, f->Size);
    		PutMsg (m);
    		break;

    	case FG_Seek:
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: do_open () - FG_Seek");
#endif
    		handle_seek (m, f);
    		break;

    	case FG_Select:
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: do_open () - FG_Select");
#endif
    		ErrorMsg (m, e & Flags_Mode);
    		break;

    	default:
#ifdef ROMSERVER_DEBUG
    		IOdebug ("rom: do_open () - default");
#endif
    		ErrorMsg (m, EC_Error + EG_FnCode + EO_File);
    		break;
    	}
    	Signal (&(f->Lock));
    }

    f->Account--;
    FreePort (reqport);

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: do_open () - THE END");
#endif
}

static void do_delete (ServInfo *	servinfo)
{
    MCB *		m = servinfo->m;
    ObjNode *	obj = GetTarget (servinfo);
    IOCCommon *	req = (IOCCommon *)(m->Control);

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: do_delete ()");
#endif

    if (obj == NULL)
    {
    	ErrorMsg (m, EO_File);
    	return;
    }

    if (!CheckMask (req->Access.Access, AccMask_D))
    {
    	ErrorMsg (m, EC_Error + EG_Protected + EO_File);
    	return;
    }

    if (obj->Type == Type_Directory && ((DirNode *)obj)->Nentries != 0)
    {
    	ErrorMsg (m, EC_Error + EG_Delete + EO_Directory);
    	return;
    }

    Unlink (obj, FALSE);

    /*
     * To stop ServLib worker indirection through Free'd
     * target on unlock
     */
    servinfo->TargetLocked = FALSE;

    Free (obj);

    ErrorMsg (m, Err_Null);
}

static void do_serverinfo (ServInfo * servinfo)
{
    MCB *	m = servinfo->m;
    WORD 	info[4];

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: do_serverinfo ()");
#endif
    InitMCB (m, 0, m->MsgHdr.Reply, NullPort, Err_Null);

    info[0] = Type_Directory;
    info[1] = RomInfo->romsize;
    info[2] = 0;
    info[3] = 1;

    m->Data = (byte*)info;
    m->MsgHdr.DataSize = 16;
    PutMsg (m);
}


/*
 *		MAIN
 */

int main (void)
{
#ifndef ROMBASEDNUCLEUS
    Environ env;
#endif
    WORD res;

    GetRoot ()->ATWFlags = 0x87654321;

#ifndef ROMBASED

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: main () - rom disk server not in nucleus");
#endif
    (void) GetEnv (MyTask->Port, &env);

#endif /* ROMBASED */

    MachineName (ProcessorName);

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: Processor Name is %s", ProcessorName);
#endif

    if ((RomInfo = OpenRom ()) == NULL)
    {
    	IOdebug ("rom: failed to find rom disk data");

    	Exit (128);
    }

#ifdef FILEBASED
    /*
     * If the rom disk file is part of the nucleus, there is no 
     * Image Header (sysbuild removes it).
     */

    if ((res = ReadRom (RomInfo,
			(BYTE *) &(RomInfo->ihdr),
                        (WORD)(sizeof (struct ImageHdr)),
			-1)) != sizeof (struct ImageHdr))
    {
    	IOdebug ("rom: error reading image header %x", res);

    	return 1;
    }

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: Image Header = [%x, %x, %x]", 
    			RomInfo->ihdr.Magic,
    			RomInfo->ihdr.Flags,
    			RomInfo->ihdr.Size);
#endif

#endif /* ROMBASEDNUCLEUS */

    if ((res = ReadRom (RomInfo,
    		    (BYTE *) &RomInfo->modhdr,
    		    (WORD)(sizeof (struct Module)),
    		    -1)) != sizeof (struct Module))
    {
    	IOdebug ("rom: error reading module header %x", res);

    	return 1;
    }
#ifdef ROMSERVER_DEBUG
    else
    {
    	IOdebug ("rom: read Module Header (%d bytes)", sizeof (struct Module));
    	IOdebug ("rom: Module Type    = %x", RomInfo->modhdr.Type);
    	IOdebug ("rom: Module Name    = %s", RomInfo->modhdr.Name);
	IOdebug ("rom: Module Version = %x", RomInfo->modhdr.Version);
    }
#endif

    if ((res = ReadRom (RomInfo,
    		    (BYTE *) &Root,
    		    (WORD)(sizeof (DirNode)),
    		    -1)) != sizeof (DirNode))
    {
    	IOdebug ("rom: error reading root DirNode %x", res);

    	return 1;
    }
#ifdef ROMSERVER_DEBUG
    else
    {
    	IOdebug ("rom: read root DirNode (%d bytes)", sizeof (struct DirNode));
    	IOdebug ("rom: Root.Name = %s", Root.Name);
    	IOdebug ("rom: Root.Entries = %d", Root.Entries);
    }
#endif
    
    RomInfo->romsize = 8;

    {
    	WORD 		number = Root.Nentries;
    	LinkNode *	parent;
    	Object *	me = Locate (Null(Object), ProcessorName);
#ifdef ROMBASED
    	WORD	rompos;
#endif
    	parent = (LinkNode *)Malloc (sizeof (LinkNode) + strlen (ProcessorName));

    	InitNode(&parent->ObjNode,
    		 "..",
    		 Type_Link,
    		 0,
    		 DefDirMatrix);

    	parent->Cap = 	me->Access;
    	strcpy (parent->Link, ProcessorName);

    	Close (me);

    	InitNode ((ObjNode *)&Root,
    		  Root.Name,
    	    	  Root.Type,
    	    	  Root.Flags,
    	    	  Root.Matrix);

    	InitList (&(Root.Entries));

    	Root.Parent = (DirNode *)parent;


    	rompos = TellRom ();

    	extract_files (&Root, number);

    	SeekRom (rompos);

    	if (call_extract_executables (number) == 0)
    	{
    	    IOdebug ("Fork () of extract_executables () failed");

    	    Exit (256);
    	}
    }

#ifdef USE_getrombuffer
    /* Read the buffer now */
    getrombuffer (RomInfo);
#endif

    RomDispatchInfo.ReqPort = NewPort ();

    {
    	NameInfo 	name;
    	Object *	ThisProcessor;
    	Object *	NameEntry;

    	ThisProcessor = Locate (Null(Object), ProcessorName);

    	if (ThisProcessor == Null(Object))
    	{
    		IOdebug ("rom : failed to locate own processor %s",
    		    ProcessorName);

    		return 1;
    	}

    	name.Port     = RomDispatchInfo.ReqPort;
    	name.Flags    = Flags_StripName;
    	name.Matrix   = DefDirMatrix;	/* BLV - for now */
    	name.LoadData = Null(WORD);

    	NameEntry = Create (ThisProcessor,
    			    Root.Name,
    			    Type_Name,
    			    sizeof (NameInfo),
    			    (BYTE *)&name);

    	if (NameEntry == Null(Object))
    	{
    		IOdebug ("rom : failed to enter name in name table, error code 0x%x",
    		    Result2(ThisProcessor));

    		return 1;
    	}
    	Close (ThisProcessor);
    }

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: dispatching ROMdisk server ... ");
#endif

    Dispatch (&RomDispatchInfo);

    CloseRom (RomInfo);

#ifdef ROMSERVER_DEBUG
    IOdebug ("rom: ROMdisk server closing down");
#endif
}
