/*{{{  Header */

/*------------------------------------------------------------------------
--                                                                      --
--                     H E L I O S			                --
--                     -----------                                      --
--                                                                      --
--             Copyright (C) 1990, Perihelion Software Ltd.             --
--         Copyright (c) 1994, Perihelion Distributed Software          --
--                        All Rights Reserved.                          --
--                                                                      --
-- eventdev.c								--
--                                                                      --
--	Interrupt based device driver for the raw keyboard server	--
--                                                                      --
--	Author:  BLV 21.2.90						--
--                                                                      --
------------------------------------------------------------------------*/

/*}}}*/
/*{{{  Includes */

#include <helios.h>
#include <queue.h>
#include <thread.h>
#include <intr.h>
#include <codes.h>
#include <device.h>
#include <stdlib.h>

#include "keyboard.h"

/*}}}*/
/*{{{  Pragmas & Constants */

#pragma -s1    	/* Suppress stack checking	*/

#define INTR_KEYBOARD	    0

/*}}}*/
/*{{{  Functions */

/*{{{  InterruptHandler */

static bool
InterruptHandler(
		 void * pDCB,
		 int	iId )
{
  HardenedSignal( &(((KeyboardDCB *)pDCB)->Device.Intr.wait) );
  return TRUE;
}

/*}}}*/
/*{{{  KeyboardMonitor */

static void
KeyboardMonitor( KeyboardDCB * pDCB )
{
  forever
    {
      BYTE key;


      Wait( &(pDCB->Device.Intr.wait) );
      
      key = *((BYTE *) 0x00006000);
      
      pDCB->ReceiveKey( (key & 0x80) == 0, key & 0x007F );
    }
}

/*}}}*/
/*{{{  StartupHardware */

static word
StartupHardware( KeyboardDCB * pDCB )
{
  IntrHandler *	pHandler = &(pDCB->Device.Intr.handler);

  
  InitSemaphore( &(pDCB->Device.Intr.wait), 0 );
  
  pHandler->Pri	 = StandardPri;
  pHandler->Code = InterruptHandler;
  pHandler->Data = pDCB;
  pHandler->Id   = INTR_KEYBOARD;

  if (SetIntrHandler( pHandler ) < Err_Null)
    return (EC_Error | EG_UnknownError | EO_Processor);

  if (! Fork( 1000, KeyboardMonitor, sizeof pDCB, pDCB ))
   {
     RemIntrHandler( pHandler );

     return (EC_Error | EG_NoMemory | EO_Processor);
   }

  return Err_Null;
}

/*}}}*/
/*{{{  ShutdownHardware */

static void
ShutdownHardware( KeyboardDCB * pDCB )
{
  RemIntrHandler( &(pDCB->Device.Intr.handler) );
}

/*}}}*/

/*{{{  DeviceClose */

static word
DeviceClose( KeyboardDCB * pDCB )
{
  ShutdownHardware( pDCB );

  return Err_Null;
}

/*}}}*/
/*{{{  DeviceOperate */

static void
DeviceOperate(
	      KeyboardDCB *	pDCB,
	      VoidFnPtr		pFn )
{
  if (pDCB->ReceiveKey != NULL)
   {
#ifdef DEBUG
     char buff[ 4 ];
     SetString_( buff, 0, 'K', ':', '!', '\0' );     
     IOdebug( buff );     
#endif
     pDCB->rc = (EC_Error | EG_InUse | EO_Processor);
     return;
   } 

  pDCB->ReceiveKey = pFn;
  pDCB->rc = StartupHardware( pDCB );
  return;
}  

/*}}}*/
/*{{{  DevOpen */

KeyboardDCB *
DevOpen(
	MPtr 	mDev,
	void *	pInfo )
{
  KeyboardDCB *	pDCB	= (KeyboardDCB *) Malloc( sizeof (KeyboardDCB) );

  
  pDCB->DCB.Device	= mDev;
  pDCB->DCB.Operate	= DeviceOperate;
  pDCB->DCB.Close	= DeviceClose;
  pDCB->ReceiveKey	= NULL;
  
  return pDCB;
}

/*}}}*/

/*}}}*/
