/*{{{  Header */

/*------------------------------------------------------------------------
--                                                                      --
--                      H E L I O S  C O M M A N D                      --
--                      --------------------------                      --
--                                                                      --
--                Copyright (C) 1993, Perihelion Software Ltd.          --
--         Copyright (C) 1994, Perihelion Distributed Software Ltd.     --
--                        All Rights Reserved.                          --
--                                                                      --
-- threadps.c								--
--                                                                      --
-- Print information about the current set of threads in the system.	--
-- Does not use any floating point code					--
--                                                                      --
--                                                                      --
-- This command will only work for processors with the generic		--
-- executive.								--
--                                                                      --
-- Author: PAB 23/4/92							--
--                                                                      --
------------------------------------------------------------------------*/
/* RcsId: $Id: threadps.c,v 1.12 1995/09/07 16:54:23 al Exp $ */

/*}}}*/
/*{{{  Includes */

#define _ID_		/* prevent link.h from defining dummy Id structure */

#include <message.h>
#include <root.h>
#include <task.h>
#include <cpustate.h>
#include <thread.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslib.h>
#include <nonansi.h>
#include <string.h>
#include <ctype.h>
#include <syslib.h>
#include <module.h>
#include <root.h>
#include <intr.h>
#include <time.h>
#include <ctype.h>

/*}}}*/
/*{{{  Types */

/* *Warning*: STOLEN from servlib.h due to incompatible headers: */
/* (kernel.h vs servlib.h)*/

typedef struct ObjNode
  {
	Node		Node;		/* link in directory list	*/
	char		Name[NameMax];	/* entry name			*/
	word		Type;		/* entry type			*/
	word		Flags;		/* flag word			*/
	Matrix		Matrix;		/* access matrix		*/
	Semaphore	Lock;		/* locking semaphore		*/
	Key		Key;		/* protection key		*/
	struct DirNode  *Parent;	/* parent directory		*/
	DateSet		Dates;		/* dates of object		*/
	word		Account;	/* owning account		*/
	word		Size;		/* object size			*/
	List		Contents;	/* whatever this object contains*/
					/* may be cast to something else*/
      }
ObjNode ;


typedef struct ThreadInfo
  {
    VoidFnPtr   ptr;			/* Added to stop segmentation faults if	*/
					/* thread terminates inbetween reading	*/
					/* the info and trying to use it	*/
    word	ss;
    word	pc;
    word	mt;
    word	sp;
    word	wakeup;
    word	status;
    word	pri;
    word	cputotal;
    word	timestart;
  }
ThreadInfo;

typedef struct Id
  {
    Node	Node;		/* For queuing on a kernel object.	*/
    word	rc;		/* result value.			*/
    SaveState * state;		/* saved process state.			*/
    word	endtime;	/* termination timeout			*/
    MCB	*	mcb;		/* MCB if present			*/
    void *	spare;		/* spare field if needed		*/
  }
Id;

/*}}}*/
/*{{{  Constants */

#define TQSTORESIZE	100
#define	RQSTORESIZE	1000
#define	KTQSTORESIZE	1000
#define FnNameMax	16	/* max length of fn name allowed */

/*}}}*/
/*{{{  Macros */

#define HeadNode_( type, pList, field )\
  (type *)((char *)((pList)->Head) - offsetof( type, field ))
    
#define NextNode_( type, pStruct, field )\
  (type *)((char *)((pStruct)->field.Next) - offsetof( type, field ))
    
#define LastNode_( pStruct, field )\
  ((pStruct)->field.Next == NULL)

/*}}}*/
/*{{{  Variables */

ExecInfo	sExecInfo;

Semaphore capture;
Semaphore display;

char	*selectname;
word	rate 		= OneSec; /* capture rate */
word	verbose		= FALSE;
word	all		= TRUE;
word	selectone	= FALSE;
word	TQCaptureTime	= 0;
word	TQCaptureTime2	= 0;
word	TQSize		= 0;
word	RQSize		= 0;
word	ResumeQSize	= 0;
word	KTQSize		= 0;


ThreadInfo TQStore[  TQSTORESIZE ];
ThreadInfo RQStore[  RQSTORESIZE ];
ThreadInfo ResumeQStore[  RQSTORESIZE ];
ThreadInfo KTQStore[ KTQSTORESIZE ];

/*}}}*/

/*{{{  Code */

/*{{{  StatusText() */

const char *
StatusText( word iStatus )
{
  static char 	aStatus[ 256 ];		/* XXX arbitary size for array */


  aStatus[ 0 ] = '\0';
  
  while (iStatus)			/* loop on status bits */
    {
      word	iBit;
      

      iBit = iStatus & (-iStatus); 	/* get least significant bit in status */

      iStatus &= ~iBit;			/* remove this bit from status */

      switch (iBit)			/* add text based on this bit */
	{
	case THREAD_SLICED:	strcat( aStatus, "Sliced" ); break;
	case THREAD_RUNNABLE:	strcat( aStatus, "Runnable" ); break;
	case THREAD_RUNNING:	strcat( aStatus, "Running" ); break;
	case THREAD_SLEEP:	strcat( aStatus, "Sleeping" ); break;
	case THREAD_SEMAPHORE:	strcat( aStatus, "Waiting on Semaphore" ); break;
	case THREAD_MSGREAD:	strcat( aStatus, "Waiting in GetMsg" ); break;
	case THREAD_MSGWRITE:	strcat( aStatus, "Waiting in PutMsg" ); break;
	case THREAD_MULTIWAIT:	strcat( aStatus, "MultiWaiting" ); break;
	case THREAD_MUTEX:	strcat( aStatus, "Waiting on Mutex" ); break;
	case THREAD_GRABMUTEX:	strcat( aStatus, "Waiting to Grab Mutexes" ); break;
	case THREAD_SAVED:	strcat( aStatus, "Saved" ); break;
	case THREAD_RESUMED:	strcat( aStatus, "Resumed" ); break;
	case THREAD_KILLED:	strcat( aStatus, "Killed" ); break;
	case THREAD_LINKWRITEQ:	strcat( aStatus, "Link Write Q" ); break;
	case THREAD_LINKREADQ:	strcat( aStatus, "Link Read Q" ); break;
	case THREAD_LINKRX:	strcat( aStatus, "LinkRx" ); break;
	case THREAD_LINKTX:	strcat( aStatus, "LinkTx" ); break;
	case THREAD_LINKWAIT:	strcat( aStatus, "LinkWait" ); break;
	case THREAD_LINKXOFF:	strcat( aStatus, "LinkXOff" ); break;
	case THREAD_LINKBUF:	strcat( aStatus, "LinkBuf" ); break;
	case THREAD_REAPED:	strcat( aStatus, "Reaped" ); break;
        case THREAD_DEBUGME:    strcat( aStatus, "DEBUG" ); break;
        case THREAD_DEBUGGER:   strcat( aStatus, "Held by debugger" ); break;
        case THREAD_DEBUGASYNC: strcat( aStatus, "Executing asynchronously" ); break;
        case THREAD_DEBUGBREAK: strcat( aStatus, "Stopped on breakpoint" ); break;
	default:		sprintf( aStatus, "<unknown %lx %lx>", iBit, iStatus );
	  return aStatus;
	}
      
      if (iStatus)
 	strcat( aStatus, ", " ); 	/* if we are going to loop, then add a comma */
    }

  return aStatus;			/* return composite status string */
}
  
/*}}}*/
/*{{{  VerboseDumpTQ() */

void
VerboseDumpTQ( void )
{
  int 	i;
  Id *	pId;

  WaitMutex( sExecInfo.TimerQLock );
  
#ifdef __ARM
  ARM_SwitchMode( SVC32Mode );
#endif

  pId = Head_( Id, *sExecInfo.TimerQ );
  
  TQCaptureTime = Timer();
  
  for (i = 0;
       !EndOfList_( pId ) && i < TQSTORESIZE;
       pId = Next_( Id, pId ), i++
       )
    {
      SaveState *	s = pId->state;

      TQStore[i].ptr       = s->InitialFn;
      TQStore[i].ss        = (word) CtoM_( s );
#ifdef __C40
      TQStore[i].pc        = s->CPUcontext.PC;
      TQStore[i].sp        = s->CPUcontext.R_USP;
#else
      TQStore[i].pc        = s->CPUcontext.R_PC;
#ifdef __ARM6
      TQStore[i].sp        = s->CPUcontext.R_USR_SP;
#else
      TQStore[i].sp        = s->CPUcontext.R_USER_SP;
#endif
#endif
      TQStore[i].mt        = s->CPUcontext.R_MT;
      TQStore[i].wakeup    = pId->endtime;
      TQStore[i].pri       = s->priority;
      TQStore[i].cputotal  = s->CPUTimeTotal;
      TQStore[i].timestart = s->InitialTime;
    }
  
  SignalMutex( sExecInfo.TimerQLock );
  
  TQSize = i;
}

/*}}}*/
/*{{{  NormalDumpTQ() */

void
NormalDumpTQ( void )
{
  int 		i;
  Id * 		pId;


  WaitMutex( sExecInfo.TimerQLock );
  
  pId = Head_( Id, *sExecInfo.TimerQ );
  
  TQCaptureTime = Timer();
  
  for (i = 0;
       !EndOfList_( pId )   && i < TQSTORESIZE;
       pId = Next_( Id, pId ), i++
       )
    {
      TQStore[i].ptr    = pId->state->InitialFn;
      TQStore[i].ss     = (word) CtoM_( pId->state );
      TQStore[i].mt     = pId->state->CPUcontext.R_MT;
      TQStore[i].wakeup = pId->endtime;
    }

  SignalMutex( sExecInfo.TimerQLock );
  
  TQSize = i;
}

/*}}}*/
/*{{{  VerboseDumpRQ() */

void
VerboseDumpRQ( word range )
{
  int 		i = 0;
  word		j;
  SaveState *	s;

  
  for (j = range; j--;)
    {
      s = Head_( SaveState, sExecInfo.RunQueues[ j ] );
      
      while (!EndOfList_( s ))
	{
	  if (i >= RQSTORESIZE)
	    break;

	  RQStore[i].ptr       = s->InitialFn;
	  RQStore[i].ss        = (word) CtoM_( s );
#ifdef __C40
	  RQStore[i].pc        = s->CPUcontext.PC;
	  RQStore[i].sp        = s->CPUcontext.R_USP;
#else
	  RQStore[i].pc        = s->CPUcontext.R_PC;
#ifdef __ARM6
	  RQStore[i].sp        = s->CPUcontext.R_USR_SP;
#else
	  RQStore[i].sp        = s->CPUcontext.R_USER_SP;
#endif
#endif
	  RQStore[i].mt        = s->CPUcontext.R_MT;
	  RQStore[i].pri       = j;
	  RQStore[i].status    = s->status;
	  RQStore[i].cputotal  = s->CPUTimeTotal;
	  RQStore[i].timestart = s->InitialTime;
	  
	  i++;

	  s = Next_( SaveState, s );
	}
    }
  
  RQSize = i;
}

/*}}}*/
/*{{{  NormalDumpRQ() */

void
NormalDumpRQ( word range )
{
  int 		i = 0;
  word		j;
  SaveState *	s;
  
  for (j = range; j--;)
    {
      s = Head_( SaveState, sExecInfo.RunQueues[ j ] );
      
      while ( !EndOfList_( s ) )
	{
	  if (i >= RQSTORESIZE)
	    break;

	  RQStore[i].ptr       = s->InitialFn;	  
	  RQStore[i].mt        = s->CPUcontext.R_MT;
	  RQStore[i].status    = s->status;
#ifdef __C40
	  RQStore[i].ss        = C40WordAddress( s );
#else
	  RQStore[i].ss        = (word)s;
#endif
	  i++;

	  s = Next_( SaveState, s );
	}
    }
  
  RQSize = i;
}

/*}}}*/
/*{{{  VerboseDumpResumeQ() */

void
VerboseDumpResumeQ( word range )
{
  int 		i = 0;
  word		j;
  SaveState *	s;

  
  for (j = range; j--;)
    {
      s = Head_( SaveState, sExecInfo.ResumeQueues[ j ] );
      
      while (!EndOfList_( s ))
	{
	  if (i >= RQSTORESIZE)
	    break;

	  ResumeQStore[i].ptr       = s->InitialFn;	  
	  ResumeQStore[i].ss        = (word) CtoM_( s );
#ifdef __C40
	  ResumeQStore[i].pc        = s->CPUcontext.PC;
	  ResumeQStore[i].sp        = s->CPUcontext.R_USP;
#else
	  ResumeQStore[i].pc        = s->CPUcontext.R_PC;
#ifdef __ARM6
	  ResumeQStore[i].sp        = s->CPUcontext.R_USR_SP;
#else
	  ResumeQStore[i].sp        = s->CPUcontext.R_USER_SP;
#endif
#endif
	  ResumeQStore[i].mt        = s->CPUcontext.R_MT;
	  ResumeQStore[i].pri       = j;
	  ResumeQStore[i].status    = s->status;
	  ResumeQStore[i].cputotal  = s->CPUTimeTotal;
	  ResumeQStore[i].timestart = s->InitialTime;
	  
	  i++;

	  s = Next_( SaveState, s );
	}
    }
  
  ResumeQSize = i;
}

/*}}}*/
/*{{{  NormalDumpResumeQ() */

void
NormalDumpResumeQ( word range )
{
  int 		i = 0;
  word		j;
  SaveState *	s;

  for (j = range; j--;)
    {
      s = Head_( SaveState, sExecInfo.ResumeQueues[ j ] );
      
      while (!EndOfList_( s ))
	{
	  if (i >= RQSTORESIZE)
	    break;
	  
	  ResumeQStore[i].ptr       = s->InitialFn;
	  ResumeQStore[i].mt        = s->CPUcontext.R_MT;
	  ResumeQStore[i].status    = s->status;
#ifdef __C40
	  ResumeQStore[i].ss        = C40WordAddress( s );
#else
	  ResumeQStore[i].ss        = (word)s;
#endif
	  i++;

	  s = Next_( SaveState, s );
	}
    }
  
  ResumeQSize = i;
}

/*}}}*/
/*{{{  VerboseDumpKTQ() */

/* Dump all known threads and their status from known thread Q */
void
VerboseDumpKTQ( void )
{
  SaveState * s; 
  int i;


  WaitMutex( sExecInfo.KnownThreadsLock );

#ifdef __ARM
  ARM_SwitchMode( SVC32Mode );
#endif
  
  for (i = 0,              s = HeadNode_( SaveState, sExecInfo.KnownThreads, threadlist );
       i < KTQSTORESIZE && ! LastNode_( s, threadlist );
       i++,                s = NextNode_( SaveState, s, threadlist )
       )
    {
      KTQStore[i].ptr       = s->InitialFn;
      KTQStore[i].ss        = (word) CtoM_( s );
#ifdef __C40
      KTQStore[i].pc        = s->CPUcontext.R_LR;
      KTQStore[i].sp        = s->CPUcontext.R_USP;
#else
#ifdef __ARM6
      KTQStore[i].pc        = s->CPUcontext.R_USR_LR;
      KTQStore[i].sp        = s->CPUcontext.R_USR_SP;
#else
      KTQStore[i].pc        = s->CPUcontext.R_USER_LR;
      KTQStore[i].sp        = s->CPUcontext.R_USER_SP;
#endif
#endif
      KTQStore[i].mt        = s->CPUcontext.R_MT;
      KTQStore[i].pri       = s->priority;
      KTQStore[i].status    = s->status;
      KTQStore[i].cputotal  = s->CPUTimeTotal;
      KTQStore[i].timestart = s->InitialTime;
    }

  KTQSize = i;

#ifdef __ARM
  ARM_SwitchMode( USR32Mode );
#endif

  SignalMutex( sExecInfo.KnownThreadsLock );
  
  return;
}

/*}}}*/
/*{{{  NormalDumpKTQ() */

void
NormalDumpKTQ( void )
{
  SaveState *	pThread; 
  int		i;


  WaitMutex( sExecInfo.KnownThreadsLock );

  for (
       i = 0,              pThread = HeadNode_( SaveState, sExecInfo.KnownThreads, threadlist );
       i < KTQSTORESIZE && !LastNode_( pThread, threadlist );
       i++,                pThread = NextNode_( SaveState, pThread, threadlist )
       )
    {
      KTQStore[ i ].ptr    = pThread->InitialFn;
      KTQStore[ i ].ss     = (word) CtoM_( pThread );
      KTQStore[ i ].mt     = pThread->CPUcontext.R_MT;
      KTQStore[ i ].status = pThread->status;
    }
  
  SignalMutex( sExecInfo.KnownThreadsLock );
  
  KTQSize = i;

  return;
}

/*}}}*/
/*{{{  TaskName() */

/* get validated taskname from mt reg. */
/* and pad it to FnNameMax chars       */
bool
TaskName(
	 word *	mt,
	 char *	name )
{
  word *	km;
  int		i;
  ObjNode *	objnode;
  Task *	tp;
  char *	tn;
  Pool *	syspool;
  Pool *	freepool;

  
  if (mt == NULL)
    {
      strcpy( name, "<mt> is NULL!  " );
      return false;
    }
  
#ifdef __C40
  mt = (word *)C40CAddress((word)mt);
#endif
  
  syspool  = &(GetRoot()->SysPool);
  freepool = GetRoot()->FreePool;
  
  if (InPool( (void *)mt, syspool ))
    {
      strcpy(name, "System");
      strncat(name, "                ", FnNameMax - 6);
      return false;
    }
  
  if (InPool( (void *)mt, freepool) )
    {
      strcpy(name, "<ModTab is Free>");
      return false;
    }
  
  if (*mt != (word) mt)
    {
      /* check its a module table pointer */
      strcpy(name, "<Invalid ModTab>");
      return false;
    }
  
#ifndef __TRAN /* Split module table  version */
  km = (word *) *(mt + 2); /* kernel module */
  
  tp = (Task *) *(km); /* data entry 0 points at task struct */
#else
  km = (word *) *(mt + 1); /* kernel module */
  
  tp = (Task *) *(km + 48); /* entry 48 points at task struct */
#endif
  
  /* get first element of taskentry */
  objnode = (ObjNode *) tp->TaskEntry;
  
  tn = objnode->Name;
  
  for (i =0 ; i < NameMax; i++)
    {
      if (tn[i] == '\0')
	{
	  strcpy(name, tn);
	  strncat(name, "                ", FnNameMax-i);
	  return true;
	}
      
      if (!isprint(tn[i]))
	{
#if 1
	  /* tn[i]=' '; */
	  sprintf( name, "0x%*.*p", -(FnNameMax - 3), FnNameMax - 3, tn );
#else
	  strcpy(name, "Corrupt name    ");
	  return false;
#endif
	}
    }
  
  strcpy(name, "<Invalid name>  ");
  return false;
  
} /* TaskName */

/*}}}*/

#ifdef __ARM
/*{{{  EmbeddedName() */

char *
EmbeddedName( VoidFnPtr pFn )
{
  word w, i;
  word *  fn;

  
  fn = (word *)((word)pFn & ~0xfc000003);	/* remove any flags and word align */

  if (fn == NULL)
    return NULL;

  /* Search up to 10 words before the STM looking for */
  /* the marker that shows where the function name is. */
  
  for (i = 0; (i < 10); i++)
    {
      w = *--fn;

      if ((w & 0xffff0000) == 0xff000000)
	{
	  return (char *)fn - (w & 0xffff);
	}
    }
  
  return NULL;
}

/*}}}*/
#endif /* __ARM */

/*{{{  FnName() */

void
FnName(
       VoidFnPtr 	fn,
       const char *	name )
{
  char *	pName =
#if defined __ARM
	      EmbeddedName( fn )
#else
	      procname( fn )
#endif
    ;

  
#ifdef __ARM
  if (pName == NULL)
    pName = procname( fn );
#endif

  if (pName == NULL)
    {
      sprintf( (char *) name, "0x%*.*p", -(FnNameMax - 3), FnNameMax - 3, fn );
      return;
    }
  
  if (sprintf( (char *) name, "%*.*s", -(FnNameMax - 1), FnNameMax - 1, pName ))
    {
      int	i;
      for (i = strlen( name ); i--;)
	if (!isprint( name[ i ] ))
	  {
	    sprintf( (char *) name, "0x%*.*p", -(FnNameMax - 3), FnNameMax - 3, fn );
	    break;
	  }
    }
  else
    ((char *)name)[ 0 ] ='\0';

  return;
}

/*}}}*/
/*{{{  ShowQs() */

void
ShowQs( bool repeat )
{
  int		i;
  uword		wakeup;
  const char *	status;
  char		taskname[ NameMax ];
  char		threadname[ FnNameMax ];

again2:  
      Wait( &display );
      
      wakeup = 0; /* I do not know the real top of Q wakeup time ! */
      
/*{{{  Timer Q */

      printf("\fTimer Queue:\n");
      
      if (verbose)
	{
	  printf("Number of sleeping threads: %ld\n",TQSize);
	  printf("Timer value at capture time: %#8lx\n", TQCaptureTime);
	}
      
      for (i = 0; i < TQSize; i++)
	{
	  TaskName((word *)(TQStore[i].mt), taskname);
	  
	  if (selectone)
	    {
	      if (strncmp( taskname, selectname, strlen(selectname) ) != 0)
		continue;
	    }
	  
	  if (!all)
	    {
	      /*
	       * @@@ Should also knock out signal handler processes
	       * but this is difficult without any names.
	       */
	      
	      if (strncmp( taskname, "ProcMan.0", 9 ) == 0 ||
		  strncmp( taskname, "System", 6 )    == 0  )
		continue;
	    }			
#ifdef __C40
	  FnName(((SaveState *)C40CAddress(TQStore[i].ss))->InitialFn, threadname);
#else
	  FnName(TQStore[i].ptr, threadname);
#endif
	  if (verbose)
	    {
	      printf( " %s: %s\n",
		     taskname, threadname);
#if 0 /* fixed C fp vararg bug? */
	      printf("\t\twakeup @ %#08lxuS, wakeup in %8lx uS (%d secs)\n",
		     TQStore[i].wakeup,
		     TQStore[i].wakeup - TQCaptureTime,
		     ((float)(TQStore[i].wakeup - TQCaptureTime))/OneSec);
#else
	      printf("\t\twakeup @ %#08lxuS, ",
		     TQStore[i].wakeup);
	      printf("wakeup in %8lx uS (%ld secs)\n",
		     TQStore[i].wakeup - TQCaptureTime,
		     (word)(TQStore[i].wakeup - TQCaptureTime) / OneSec);
#endif
	      printf("\t\trun time %8dS, CPU time consumed %ldS\n",
		     time(NULL) - (time_t)TQStore[i].timestart,
		     (TQStore[i].cputotal) / OneSec);
	      
	      printf("\t\tss %8lx, physpri %1lx, pc %8lx, mt %8lx, usp %8lx\n",
		     TQStore[i].ss, TQStore[i].pri,
		     TQStore[i].pc, TQStore[i].mt,
		     TQStore[i].sp);
	    }
	  else
	    {
	      printf( " %s: %s: wakeup in %ld secs\n",
		     taskname, threadname,
		     (word)(TQStore[i].wakeup - TQCaptureTime) / OneSec);
	    }
	}
      

/*}}}*/
/*{{{  Run Qs */

      printf("Run Queues:\n");
      
      if (verbose)
	printf("Number of runnable threads = %ld\n", RQSize);
      
      for (i=0; i < RQSize; i++)
	{
	  TaskName((word *)(RQStore[i].mt),taskname);
	  
	  if (selectone)
	    {
	      if (strncmp( taskname, selectname, strlen(selectname) ) != 0)
		continue;
	    }

	  if (!all)
	    {
	      if (strncmp( taskname, "ProcMan.0", 9 ) == 0 ||
		  strncmp( taskname, "System", 6 )    == 0  )
		continue;
	    }			
#ifdef __C40
	  FnName(((SaveState *)C40CAddress(RQStore[i].ss))->InitialFn, threadname);
#else
	  FnName(RQStore[i].ptr, threadname);
#endif
	  status = StatusText(RQStore[i].status);
	  
	  if (verbose)
	    {
	      printf( " %s: %s: %s\n", taskname, threadname, status);
	      
	      printf("\t\trun time %8dS, CPU time consumed %ldS\n",
		     time(NULL) - (time_t)RQStore[i].timestart,
		     RQStore[i].cputotal / OneSec);
	      
	      printf("\t\tss %8lx, physpri %1lx, pc %8lx, mt %8lx, usp %8lx\n",
		     RQStore[i].ss, RQStore[i].pri,
		     RQStore[i].pc, RQStore[i].mt,
		     RQStore[i].sp);
	    } else
	      printf( " %s: %s: %s\n", taskname, threadname, status);
	}
      

/*}}}*/
/*{{{  Resume Qs */

      printf("Resume Queues:\n");
      
      if (verbose)
	printf("Number of resumable threads = %ld, HighestAvailPri = %ld\n",
	       ResumeQSize, *sExecInfo.HighestAvailPri );
      
      for (i=0; i < ResumeQSize; i++)
	{
	  TaskName((word *)(ResumeQStore[i].mt),taskname);
	  
	  if (selectone)
	    {
	      if (strncmp( taskname, selectname, strlen(selectname) ) != 0)
		continue;
	    }

	  if (!all)
	    {
	      if (strncmp( taskname, "ProcMan.0", 9 ) == 0 ||
		  strncmp( taskname, "System", 6 )    == 0  )
		continue;
	    }			
#ifdef __C40
	  FnName(((SaveState *)C40CAddress(ResumeQStore[i].ss))->InitialFn, threadname);
#else
	  FnName(ResumeQStore[i].ptr, threadname);
#endif
	  status = StatusText(ResumeQStore[i].status);
	  
	  if (verbose)
	    {
	      printf( " %s: %s: %s\n", taskname, threadname, status);
	      
	      printf("\t\trun time %8dS, CPU time consumed %ldS\n",
		     time(NULL) - (time_t)ResumeQStore[i].timestart,
		     ResumeQStore[i].cputotal / OneSec);
	      
	      printf("\t\tss %8lx, physpri %1lx, pc %8lx, mt %8lx, usp %8lx\n",
		     ResumeQStore[i].ss, ResumeQStore[i].pri,
		     ResumeQStore[i].pc, ResumeQStore[i].mt,
		     ResumeQStore[i].sp);
	    } else
	      printf( " %s: %s: %s\n", taskname, threadname, status);
	}
      

/*}}}*/
/*{{{  Known Threads */

      printf("All threads:\n");
      
      if (verbose)
	{
	  printf("Total number of threads: %ld\n", KTQSize);
	}
      
      /* print known thread Q contents */
      
      for (i=0; i < KTQSize; i++)
	{
	  TaskName((word *)(KTQStore[i].mt),taskname);
	  
	  if (selectone)
	    {
	      if (strncmp( taskname, selectname, strlen(selectname) ) != 0)
		continue;
	    }

	  if (!all)
	    {
	      if (strncmp( taskname, "ProcMan.0", 9 ) == 0 ||
		  strncmp( taskname, "System", 6 )    == 0  )
		continue;
	    }			
#ifdef __C40
	  FnName(((SaveState *)C40CAddress(KTQStore[i].ss))->InitialFn, threadname);
#else
	  FnName(KTQStore[i].ptr, threadname);
#endif
	  status = StatusText(KTQStore[i].status);
	  
	  printf( " %s: %s: %s\n", taskname, threadname, status);
	  
	  if (verbose)
	    {
	      printf("\t\trun time %8dS, CPU time consumed %ldS\n",
		     time(NULL) - (time_t)KTQStore[i].timestart,
		     KTQStore[i].cputotal / OneSec);
	      
	      printf("\t\tss %8lx, physpri %1lx, pc %8lx, mt %8lx, usp %8lx\n",
		     KTQStore[i].ss, KTQStore[i].pri,
		     KTQStore[i].pc, KTQStore[i].mt,
		     KTQStore[i].sp);
	    }
	}
      

/*}}}*/

  if (repeat)
    {
      Signal( &capture );
      goto again2;
    }

  return;
}

/*}}}*/
/*{{{  VerboseDumpQs() */

void
VerboseDumpQs( word range )
{

    /* Timer handler does not check TimerQ mutex, so must disable
       clock interrupts whilst looking at the timer queue  */
#ifdef __ARM
  DisableTimeouts();
#endif

  VerboseDumpTQ();

#ifdef __ARM
  EnableTimeouts();
#endif

  VerboseDumpKTQ();

  IntsOff();
  VerboseDumpRQ( range );
  VerboseDumpResumeQ( range );
  IntsOn();
}

/*}}}*/
/*{{{  NormalDumpQs() */

void
NormalDumpQs( word range )
{

    /* Timer handler does not check TimerQ mutex, so must disable
       clock interrupts whilst looking at the timer queue  */
#ifdef __ARM
  DisableTimeouts();
#endif

  NormalDumpTQ();

#ifdef __ARM
  EnableTimeouts();
#endif
  
  NormalDumpKTQ();
  
  IntsOff();
  NormalDumpRQ( range );
  NormalDumpResumeQ( range );
  IntsOn();
}

/*}}}*/
/*{{{  DumpThread() */

void
DumpThread( bool repeat )
{
  word	range = GetPhysPriRange();

  
 again:
      
  if (verbose)
    {
      VerboseDumpQs( range );
    }
  else
    {
      NormalDumpQs( range );
    }

  Signal( &display );
  
  if (repeat)
    {
      Wait( &capture );

      Delay( rate );
	  
      goto again;
    }
}

/*}}}*/
/*{{{  main() */

int
main (int argc, char **argv)
{
  bool	repeat = FALSE;

  
  GetExecInfo( &sExecInfo );
  
  InitSemaphore(&display,0);
  InitSemaphore(&capture,0);
  
  while (*(++argv) != NULL)
    {
      if (strcmp(*argv, "-r") == 0)
	repeat = TRUE;
      
      if (strcmp(*argv, "-l") == 0)
	verbose = TRUE;
      
      if (strcmp(*argv, "-s") == 0)	/* XXX-ought to be a '-a' option to enable full listing */
      {
	  if( selectone == TRUE )
	  {
	    fprintf(stderr,"Error: -s option ignored with -c option\n");
	  }
	  else
	    all = FALSE;
      }

      if (strncmp(*argv, "-c", 2) == 0)
      {
	if( all == FALSE )
	{
	  fprintf(stderr,"Error: -c option ignored with -s option\n");
	}
	else
	{
	  char* arg = *argv;
	  selectone = TRUE;
	  if( arg[2] == NULL )
	    /* Error, give help */
	    arg[1] = 'h';
	  else
	  {
	    selectname = (char *)malloc(strlen(&arg[2]));
	    strcpy(selectname, &arg[2]);
	  }
	}
      }
      
      if (strncmp(*argv, "-h", 2) == 0)
	{
	  fprintf(stderr,"usage: threadps [-r] [-l] [-s] [-h] [-c<process>]\n" );
	  fprintf(stderr,"       -r: repeat forever\n" );
	  fprintf(stderr,"       -l: long listing\n" );
	  fprintf(stderr,"       -s: ignore boring processes\n" );
	  fprintf(stderr,"       -h: display help message\n" );
	  fprintf(stderr,"       -c<process>: only show <process>'s threads\n" );
	  exit(0);		    
	}		
    }

  Fork( 2000, DumpThread, sizeof (repeat), repeat );
  
  ShowQs( repeat );

  return 0;
  
} /* main */

/*}}}*/

/*}}}*/

/* end of threadps.c */
